/*
Read a CBF_BYTE_OFFSET compressed CBF image, using only the "mini-header".
This is the byte offset decompression algorithm by Wolfgang Kabsch (originally
in FORTRAN).

This version recognizes the otherwise unused code of 0x800000 as a flag to 
reset the running value (cval) to zero.  Separate images, or subsections of
images, can then be concatenated by including this code as a separator in
the byte stream.


ObjPtr is a struct to carry the destination image information.
The other parameters in the function call are not used.

Eric F. Eikenberry 31 Aug 2009
*/


/****************************************************************************\
**                                                                          **
\****************************************************************************/
// Forced reading of a cbf image not using CBF_lib
// Only for CBF_BYTE_OFFSET or CBF_NONE compression

int read_diskimage_cbf_raw(struct object_descriptor *ObjPtr,
		unsigned long byteoffset, unsigned long numbytes, char *buffer)
{
struct stat statbuf;
char *buf, *p;
int size, *dp, diff, nerr, binsize, cval;
union {	char *cp;
		unsigned char *ucp;
		short *sp;
		unsigned short *usp;
		int *ip;
		} t;

// Header information has been gathered by get_im_head_cbf()

if (!ObjPtr->Handle)
	return -1;
if (!(size = ObjPtr->Wide*ObjPtr->High*ObjPtr->Bpp/8))
	return -1;;

(void)lseek(ObjPtr->Handle, 0L, SEEK_SET);	// rewind file 

fstat(ObjPtr->Handle, &statbuf);
if (!(buf = malloc(statbuf.st_size)))
	return -1;
nerr = read(ObjPtr->Handle, buf, statbuf.st_size);
if (nerr < 0)
	{
	free(buf);
	return -1;
	}

if (ObjPtr->Data)
	{
	free(ObjPtr->Data);
	ObjPtr->Data=NULL;
	}
if (!(ObjPtr->Data=malloc(size)))
	{
	free(buf);
	return -1;
	}

// is it uncompressed?
if ( (p=strstr(buf, "CBF_NONE")) )
	{
	if ( !(p=strstr(buf, "X-Binary-Size:")) )
		{
		printf("*** Bad header - cannot find X-Binary-Size\n");
		free(buf);
		return -1;
		}
	p += strlen("X-Binary-Size:");
	sscanf(p, "%d", &binsize);
	if ( !(p=strstr(p, "\f\032\004\325")) )
		{
		printf("*** Bad header - cannot find data start\n");
		free(buf);
		return -1;
		}
	p += strlen("\f\032\004\325");
	memcpy(ObjPtr->Data, p, binsize);
	free(buf);
	return 0;
	}

// find the start of the data
if ( !(p=strstr(buf, "CBF_BYTE_OFFSET")) )
	{
	printf("*** Bad header - cannot find CBF_BYTE_OFFSET\n");
	free(buf);
	return -1;
	}
if ( !(p=strstr(buf, "X-Binary-Size:")) )
	{
	printf("*** Bad header - cannot find X-Binary-Size\n");
	free(buf);
	return -1;
	}
p += strlen("X-Binary-Size:");
sscanf(p, "%d", &binsize);
if ( !(p=strstr(p, "\f\032\004\325")) )
	{
	printf("*** Bad header - cannot find data start\n");
	free(buf);
	return -1;
	}
p += strlen("\f\032\004\325");

dp = (int *)ObjPtr->Data;
t.cp = p;							// start of data
cval = 0;
while (t.cp-p < binsize)
	{
	/*
	** If the file is corrupted with a lot of zeroes, 
	** then (*t.ucp != 0x80) is true too many times,
	** and 'dp' tries to go beyond the buffer end.
	*/
	if ((char *)dp >= (ObjPtr->Data+size))
		break;
	if (*t.ucp != 0x80)
		diff = (int)*t.cp++;		// 1 byte diff
	else							// *t.cp == 0x80
		{
		t.cp++;						// advance 1 byte
		if (*t.sp==0)				// flag to reset cval
			{
			cval = 0;
			t.sp++;					// advance over flag
			continue;
			}
		if (*t.usp != 0x8000)
			diff = (int)*t.sp++;	// 2 byte diff
		else						// *t.sp == 0x8000
			{
			t.sp++;					// advance 2 bytes
			diff = *t.ip++;			// 4 byte diff
			}
		}
	cval = *dp++ = cval + diff;
	}

free(buf);
return 0;
}

